# -*- coding: utf-8 -*-

"""Built-in reference parsers."""


import contextlib
import os
import importlib
import pip
import tempfile
import shutil
from zipfile import ZipFile

import boto3facade
from boto3facade.s3 import S3
from boto3facade.cloudformation import Cloudformation

from humilis.exceptions import ReferenceError
import humilis.utils as utils


def _get_s3path(layer, config, full_path):
    """Returns the S3 target (bucket, key) for a local file."""
    env_prefix = "{base_prefix}{env_name}/".format(
        base_prefix=config.profile.get('s3prefix', ''),
        env_name=layer.env_name)
    if layer.env_stage is not None:
        env_prefix = "{}{}/".format(env_prefix, layer.env_stage)

    s3key = "{env_prefix}{layer_name}/{file_name}".format(
        env_prefix=env_prefix,
        layer_name=layer.name,
        file_name=os.path.basename(full_path))
    s3bucket = config.profile.get('bucket')
    return (s3bucket, s3key)


@utils.reference_parser()
def file(layer, config, path=None):
    """Uploads a local file to S3 and returns the corresponding S3 path.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param path: Path to the file, relative to the location of meta.yaml.

    :returns: The S3 path where the file has been uploaded.
    """
    full_path = os.path.join(layer.basedir, path)
    s3bucket, s3key = _get_s3path(layer, config, full_path)
    s3 = S3(config)
    s3.cp(full_path, s3bucket, s3key)
    layer.logger.info("{} -> {}/{}".format(full_path, s3bucket, s3key))
    return {'s3bucket': s3bucket, 's3key': s3key}


@utils.reference_parser(name='lambda')
def lambda_ref(layer, config, path=None):
    """Prepares a lambda deployment package and uploads it to S3.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param path: Path to the file, relative to the location of meta.yaml.

    :returns: S3 path where the deployment package has been uploaded.
    """
    full_path = os.path.abspath(os.path.join(layer.basedir, path))
    logger = layer.logger
    if os.path.isdir(full_path):
        with _make_deploy_package(full_path, logger) as full_path:
            s3path = file(layer, config, full_path)
    else:
        path_no_ext, ext = os.path.splitext(full_path)
        if ext == '.zip':
            s3path = file(layer, config, full_path)
        else:
            with _make_simple_deploy_package(full_path, logger) as full_path:
                s3path = file(layer, config, full_path)

    return s3path


@contextlib.contextmanager
def _make_deploy_package(full_path, logger):
    """Creates a deployment package for multi-file lambda with deps."""
    tmpdir = tempfile.mkdtemp()
    basename = os.path.basename(full_path)
    targetdir = os.path.join(tmpdir, basename)
    shutil.copytree(full_path, targetdir)
    setup_file = os.path.join(targetdir, 'setup.py')
    if os.path.isfile(setup_file):
        # Install all depedendencies in the same dir
        pip.main(['install', targetdir, '-t', targetdir])

    zipfile = os.path.join(tmpdir, basename + '.zip')
    with ZipFile(zipfile, 'w') as myzip:
        utils.zipdir(targetdir, myzip)
    yield zipfile
    shutil.rmtree(tmpdir)


@contextlib.contextmanager
def _make_simple_deploy_package(full_path, logger):
    """Creates a deployment package for a one-file no-deps lambda."""
    logger.info("Creating deployment package for '{}'".format(full_path))
    tmpdir = tempfile.mkdtemp()
    path_no_ext, ext = os.path.splitext(full_path)
    basename = os.path.basename(path_no_ext)
    zipfile = os.path.join(tmpdir, basename + '.zip')
    with ZipFile(zipfile, 'w') as myzip:
        myzip.write(full_path)
    yield zipfile
    shutil.rmtree(tmpdir)


@utils.reference_parser()
def layer(layer, config, layer_name=None, resource_name=None):
    """Gets the physical ID of a resource in an already deployed layer.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param layer_name: The name of the layer that contains the target resource.
    :param resource_name: The logical name of the target resource.

    :returns: The physical ID of the resource
    """
    stack_name = utils.get_cf_name(layer.env_name, layer_name,
                                   stage=layer.env_stage)
    cf = Cloudformation(config)
    resource = cf.get_stack_resource(stack_name, resource_name)

    if len(resource) < 1:
        all_stack_resources = [x.logical_resource_id for x
                               in cf.get_stack_resources(stack_name)]
        msg = "{} does not exist in stack {} (with resources {}).".format(
            resource_name, stack_name, all_stack_resources)
        raise ReferenceError(resource_name, msg, logger=layer.logger)

    return resource[0].physical_resource_id


@utils.reference_parser()
def output(layer, config, layer_name=None, output_name=None):
    """Gets the value of an output produced by an already deployed layer.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param layer_name: The logical name of the layer that produced the output.
    :param output_name: The logical name of the output parameter.
    """
    stack_name = utils.get_cf_name(layer.env_name, layer_name,
                                   stage=layer.env_stage)
    cf = Cloudformation(config)
    output = cf.get_stack_output(stack_name, output_name)
    if len(output) < 1:
        all_stack_outputs = list(cf.stack_outputs(layer_name).keys())
        msg = ("{} output does not exist for stack {} "
               "(with outputs {}).").format(output_name,
                                            stack_name,
                                            all_stack_outputs)
        ref = "output ({}/{})".format(layer_name, output_name)
        raise ReferenceError(ref, msg, logger=layer.logger)
    return output[0]


@utils.reference_parser()
def boto3(layer, config, service=None, call=None, output_attribute=None,
          output_key=None):
    """Calls a boto3facade method.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param service: The name of the AWS service.
    :param call: A dict with two keys: method, and parameters.
    :param output_attribute: Object attribute to return.
    :param output_key: Dictionary key to return.

    :returns: The call response, or its corresp. attribute or key.
    """
    facade_name = service.title()
    if not hasattr(boto3facade, service):
        ref = "boto3facade.{}.{}.{}: {}".format(service, facade_name,
                                                call['method'],
                                                call['parameters'])
        msg = "Service {} not supported".format(service)
        raise ReferenceError(ref, msg, logger=layer.logger)

    module = importlib.import_module("boto3facade.{}".format(service))
    facade_cls = getattr(module, facade_name)
    facade = facade_cls(config)
    method = getattr(facade, call['method'])
    args = call.get('args', [])
    kwargs = call.get('kwargs', {})
    result = method(*args, **kwargs)
    # If the result is a sequence, we return just the first item
    if hasattr(result, '__iter__'):
        result = list(result)[0]

    if output_attribute is not None:
        return getattr(result, output_attribute)
    elif output_key is not None:
        return result.get(output_key)
    else:
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import abc
import logging
import os
import io
import glob
import json
from sys import exit

import yaml
import jinja2

from humilis.exceptions import FileFormatError


def zipdir(path, ziph):
    """Adds all files under a path to a zip file."""
    for root, dirs, files in os.walk(path):
        for file in files:
            relroot = os.path.relpath(root, path)
            arcname = os.path.join(relroot, file)
            ziph.write(os.path.join(root, file), arcname=arcname)


def unroll_tags(tags):
    """Unrolls the tag list of a resource into a dictionary."""
    return {tag['Key']: tag['Value'] for tag in tags}


def roll_tags(tags):
    """Rolls a dictionary of tags into a list of tags Key/Value dicts."""
    return [{'Key': k, 'Value': v} for k, v in tags.items()]


def get_cf_name(env_name, layer_name, stage=None):
    """Produces the CF stack name for layer."""
    cf_name = "{}-{}".format(env_name, layer_name)
    if stage is not None:
        cf_name = "{}-{}".format(cf_name, stage)
    return cf_name


def reference_parser(name=None):
    def decorator(func):
        """Declares a function as a reference parser."""
        setattr(func, '__is_humilis_reference_parser__', True)
        setattr(func, '__humilis_reference_parser_name__', name)
        return func
    return decorator


class TemplateLoader:
    @abc.abstractmethod
    def load_section(self, *args, **kwargs):
        pass


class DirTreeBackedObject(TemplateLoader):
    """Loads data from a directory tree of files in various formats."""
    def __init__(self, basedir, logger=None):
        self.basedir = basedir
        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

    @abc.abstractproperty
    def params(self):
        pass

    def get_section_files(self, section):
        """
        Produces a list of all files associated with a dir tree section
        """
        # We read all files within the section dir, and merge them in a dict
        basedir = os.path.join(self.basedir, section)
        if os.path.isdir(basedir):
            # Section file(s) in their own directory
            section_files = []
            for (dirpath, dirnames, filenames) in os.walk(basedir):
                section_files += [os.path.join(dirpath, fn)
                                  for fn in filenames]
        else:
            section_files = list(glob.glob("{}/{}.*".format(self.basedir,
                                                            section)))

        return section_files

    def load_section(self, section, files=None, params={}):
        """
        Reads all files associated with a layer section (parameters, resources,
        mappings, etc)
        """
        if files is None:
            files = self.get_section_files(section)

        data = {}

        for filename in files:
            self.logger.info("Loading {}".format(filename))
            with open(filename, 'r') as f:
                this_data = self.load_file(filename, f, params=params)
            if this_data is None:
                continue

            if len(this_data) != 1:
                raise FileFormatError(filename, self.logger)

            data_key = list(this_data.keys())[0]
            if data_key.lower() != section.lower():
                self.logger.critical("Error parsing %s: %s was expected but "
                                     "%s was found" %
                                     (filename, section.title(), data_key))
                exit(1)
            for k, v in list(this_data.values())[0].items():
                data[k] = v

        return data

    def load_file(self, filename, f, params={}):
        filename, file_ext = os.path.splitext(filename)
        if file_ext in {'.yml', '.yaml'}:
            data = yaml.load(f)
        elif file_ext == '.json':
            data = json.load(f)
        elif file_ext == '.j2':
            template = jinja2.Template(f.read())
            data = template.render(**params)
            data = self.load_file(filename, io.StringIO(data))
        else:
            self.logger.critical("Error loading %s: unknown file "
                                 "extension %s" % filename, file_ext)
            exit(1)

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import pytest

import humilis.utils as utils
from humilis.exceptions import CloudformationError


def test_create_layer_object(test_environment, test_vpc_layer):
    layer = test_vpc_layer
    assert layer.name == 'vpc'
    assert layer.cf_name == utils.get_cf_name(
        layer.env_name,
        layer.name,
        stage=layer.env_stage)
    assert len(layer.yaml_params) == 2
    assert layer.yaml_params['vpc_cidr']['value'] == '10.0.0.0/16'
    assert layer.tags.get('humilis-layer') == layer.name
    assert layer.tags.get('humilis-environment') == test_environment.name


def test_get_section_files(test_vpc_layer):
    assert len(test_vpc_layer.loader.get_section_files('resources')) == 2
    assert len(test_vpc_layer.loader.get_section_files('meta')) == 1
    assert len(test_vpc_layer.loader.get_section_files('invalid')) == 0


def test_create_and_delete_layer(cf, test_roles_layer):
    # Make sure the stack wasn't there already
    assert not cf.stack_exists(test_roles_layer.cf_name)

    # Create the stack, and make sure it has been pushed to CF
    cf_template = test_roles_layer.create()
    assert isinstance(cf_template, dict)
    assert cf.stack_ok(test_roles_layer.cf_name)

    # Delete the stack
    test_roles_layer.delete()
    assert not cf.stack_exists(test_roles_layer.cf_name)


def test_load_section(test_vpc_layer):
    files = test_vpc_layer.loader.get_section_files('resources')
    data = test_vpc_layer.loader.load_section('resources', files)
    assert all(res in data for res in ['AttachGateway', 'Subnet'])


def test_compile_template(test_vpc_layer):
    cf_template = test_vpc_layer.compile()
    assert 'VPC' in cf_template['Resources'] and \
           'InternetGateway' in cf_template['Resources'] and \
           'Description' in cf_template and \
           len(cf_template['Description']) > 0


def test_create_layer_lacking_dependencies(cf, test_streams_roles_layer,
                                           test_streams_layer):
    """Attempts to create a stack lacking dependencies: exception"""
    test_streams_layer.delete()
    assert not cf.stack_exists(test_streams_roles_layer.cf_name)
    assert not cf.stack_exists(test_streams_layer.cf_name)
    # Should simply skip the layer since dependencies are not met
    test_streams_roles_layer.create()
    assert not cf.stack_exists(test_streams_roles_layer.cf_name)


def test_create_layer_absent_section_dirs(cf, test_streams_layer):
    """Attempts to create a layer without section directories"""
    test_streams_layer.create()
    assert cf.stack_exists(test_streams_layer.cf_name)


def test_create_dependant_layer(cf, test_vpc_layer, test_instance_layer):
    """Creates two stacks, the second depending on the first"""
    test_vpc_layer.create()
    assert cf.stack_ok(test_vpc_layer.cf_name)
    test_instance_layer.create()
    assert cf.stack_ok(test_instance_layer.cf_name)
    test_instance_layer.delete()
    assert not cf.stack_exists(test_instance_layer.cf_name)
    test_vpc_layer.delete()
    assert not cf.stack_exists(test_vpc_layer.cf_name)


def test_create_namedinstance_stack(cf, test_vpc_layer,
                                    test_named_instance_layer):
    """Referring to AMIs using their tags"""
    test_vpc_layer.create()
    assert cf.stack_ok(test_vpc_layer.cf_name)
    test_named_instance_layer.create()
    assert cf.stack_ok(test_named_instance_layer.cf_name)
    test_named_instance_layer.delete()
    assert not cf.stack_exists(test_named_instance_layer.cf_name)
    test_vpc_layer.delete()
    assert not cf.stack_exists(test_vpc_layer.cf_name)


def test_get_outputs_from_nondeployed_layer(cf, test_vpc_layer):
    """Tries to get outputs from a layer thas has not been deployed: error"""
    with pytest.raises(CloudformationError):
        test_vpc_layer.outputs


def test_get_outputs_from_layer_without_outputs(cf, test_vpc_layer):
    """Gets outputs from a layer without outputs"""
    test_vpc_layer.create()
    assert test_vpc_layer.outputs is None
    test_vpc_layer.delete()


def test_get_outputs_from_layer(cf, test_streams_layer):
    """Gets outputs from a layer that does produce outputs"""
    test_streams_layer.create()
    ly = test_streams_layer.outputs
    assert isinstance(ly, dict)
    # The names of the 4 Kinesis streams in the layer
    assert len(ly) == 4



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import click

from humilis.config import config
from humilis.environment import Environment

LOG_LEVELS = ['critical', 'error', 'warning', 'info', 'debug']


def validate_log_level(ctx, param, value):
    value = value.lower()
    if value not in LOG_LEVELS:
        raise click.BadParameter("Should one of {}".format(LOG_LEVELS))
    return value.upper()


@click.group()
@click.option('--log', default='info', help="Log level: {}".format(LOG_LEVELS),
              callback=validate_log_level, metavar='LEVEL')
@click.option('--profile', default='default', metavar='NAME')
def main(log, profile):
    logging.basicConfig(level=getattr(logging, log))
    config.boto_config.activate_profile(profile)


@main.command()
@click.argument('environment')
@click.option('--stage', help="Deployment stage, e.g. PRODUCTION, or DEV",
              default=None, metavar='STAGE')
@click.option('--output', help="Store environment outputs in a yaml file",
              default=None, metavar='FILE')
@click.option('--pretend/--no-pretend', default=False)
def create(environment, stage, output, pretend):
    """Creates an environment."""
    env = Environment(environment, stage=stage)
    if not pretend:
        env.create(output_file=output, update=False)


@main.command()
@click.argument('environment')
@click.option('--stage', help="Deployment stage, e.g. PRODUCTION, or DEV",
              default=None, metavar='STAGE')
@click.option('--output', help="Store environment outputs in a yaml file",
              default=None, metavar='FILE')
@click.option('--pretend/--no-pretend', default=False)
def update(environment, stage, output, pretend):
    """Updates (or creates) an environment."""
    env = Environment(environment, stage=stage)
    if not pretend:
        env.create(output_file=output, update=True)


@main.command()
@click.argument('environment')
@click.option('--stage', help="Deployment stage, e.g. PRODUCTION, or DEV",
              default=None, metavar='STAGE')
@click.option('--pretend/--no-pretend', default=False)
def delete(environment, stage, pretend):
    """Deletes an environment that has been deployed to CF."""
    env = Environment(environment, stage=stage)
    if not pretend:
        env.delete()


@main.command()
@click.option('--ask/--no-ask', default=True)
def configure(ask):
    """Configure humilis."""
    config.boto_config.configure(ask=ask)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import logging
import os

from boto3facade.cloudformation import Cloudformation
import yaml

from humilis.config import config
from humilis.exceptions import FileFormatError
from humilis.layer import Layer
import humilis.utils as utils


class Environment():
    """Manages the deployment of a collection of humilis layers."""
    def __init__(self, yml_path, logger=None, stage=None):
        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger
        self.__yml_path = yml_path
        self.name = os.path.splitext(os.path.split(yml_path)[1])[0]
        self.stage = stage and stage.upper()
        self.basedir = os.path.split(yml_path)[0]
        with open(yml_path, 'r') as f:
            self.meta = yaml.load(f).get(self.name)

        if len(self.meta) == 0:
            raise FileFormatError(yml_path, logger=self.logger)

        self.cf = Cloudformation(config.boto_config)
        self.sns_topic_arn = self.meta.get('sns-topic-arn', [])
        self.tags = self.meta.get('tags', {})
        self.tags['humilis-environment'] = self.name

        self.layers = []
        for layer in self.meta.get('layers', []):
            layer_name = layer.get('layer', None)
            if layer_name is None:
                msg = "Wrongly formatted layer: {}".format(layer)
                raise FileFormatError(yml_path, msg)
            if layer.get('disable', False):
                message = ("Layer {} is disabled by configuration. "
                           "Skipping.".format(layer.get('layer')))
                self.logger.warning(message)
                continue

            # Get the layer params provided in the environment spec
            layer_params = {k: v for k, v in layer.items() if k != 'layer'}
            layer_obj = Layer(self, layer_name, **layer_params)
            self.layers.append(layer_obj)

    @property
    def outputs(self):
        """Outputs produced by each environment layer"""
        outputs = {}
        for layer in self.layers:
            ly = layer.outputs
            if ly is not None:
                outputs[layer.name] = ly
        return outputs

    def create(self, output_file=None, update=False):
        """Creates or updates an environment."""
        self.populate_hierarchy()
        for layer in self.layers:
            layer.create(update=update)
        self.logger.info({"outputs": self.outputs})
        if output_file is not None:
            self.write_outputs(output_file)

    def write_outputs(self):
        """Writes layer outputs to a YAML file."""
        with open("{}.outputs.yaml".format(self.name), 'w') as f:
            f.write(yaml.dump(self.outputs, indent=4,
                              default_flow_style=False))

    def populate_hierarchy(self):
        """Adds tags to env layers indicating parent-child dependencies."""
        for layer in self.layers:
            if layer.depends_on and len(layer.depends_on) > 0:
                for parent_name in layer.depends_on:
                    layer = self.get_layer(parent_name).add_child(layer)

    def get_layer(self, layer_name):
        """Gets a layer by name"""
        sel_layer = [layer for layer in self.layers
                     if layer.cf_name == layer_name]
        if len(sel_layer) > 0:
            return sel_layer[0]

    def delete(self):
        """Deletes the complete environment from CF."""
        for layer in reversed(self.layers):
            layer.delete()

    @property
    def in_cf(self):
        """Returns true if the environment has been deployed to CF."""
        return self.name in {
            utils.unroll_tags(stk['Tags']).get('humilis-environment')
            for stk in self.cf.stacks}

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "Environment('{}')".format(self.__yml_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from click.testing import CliRunner
import humilis
import humilis.cli
import pytest


ENV_ACTIONS = ['create', 'delete', 'update']
LOG_LEVELS = ['critical', 'error', 'warning', 'info', 'debug']


@pytest.yield_fixture(scope="module")
def runner():
    yield CliRunner()


@pytest.mark.parametrize("action", ENV_ACTIONS)
def test_env_actions(action, runner, environment_definition_path):
    result = runner.invoke(humilis.cli.main, [action, '--pretend',
                                              environment_definition_path])
    assert result.exit_code == 0


def test_configure_action(runner):
    result = runner.invoke(humilis.cli.main, ['configure', '--no-ask'])
    assert result.exit_code == 0


def test_actions_with_missing_environment(runner):
    for action in ENV_ACTIONS:
        result = runner.invoke(humilis.cli.main, [action])
        assert result.exit_code > 0
        assert isinstance(result.exception, SystemExit)


def test_invalid_log_level(runner):
    result = runner.invoke(humilis.cli.main, ['--log', 'invalid'])
    assert result.exit_code > 0
    assert isinstance(result.exception, SystemExit)


@pytest.mark.parametrize('level', LOG_LEVELS)
def test_valid_log_level(level, runner, environment_definition_path):
    result = runner.invoke(humilis.cli.main, ['--log', level, 'create',
                                              environment_definition_path,
                                              '--pretend'])
    assert result.exit_code == 0


def test_output(runner, environment_definition_path):
    result = runner.invoke(humilis.cli.create, [environment_definition_path,
                                                '--output', 'filename',
                                                '--pretend'])
    assert result.exit_code == 0


def test_invalid_option(runner, environment_definition_path):
    result = runner.invoke(humilis.cli.create, [environment_definition_path,
                                                '--invalid_option', 'whatever'
                                                '--pretend'])
    assert result.exit_code > 0


def test_stage(runner, environment_definition_path):
    result = runner.invoke(humilis.cli.create, [environment_definition_path,
                                                '--stage', 'production',
                                                '--pretend'])
    assert result.exit_code == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import humilis.metadata as metadata
import os

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    if os.path.isfile('README.md'):
        long_description = open('README.md').read()
    else:
        long_description = metadata.description

setup(
    name=metadata.project,
    version=metadata.version,
    author=metadata.authors[0],
    author_email=metadata.emails[0],
    url=metadata.url,
    license=metadata.license,
    description=metadata.description,
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        'PyYAML',
        'Click',
        'boto3',
        'boto3facade==0.0.9',
        'jinja2'],
    classifiers=[
        "Programming Language :: Python :: 3"],
    # Allow tests to be run with `python setup.py test'.
    tests_require=[
        'pytest>=2.5.1',
        'mock>=1.0.1',
        'flake8>=2.1.0'
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'humilis=humilis.cli:main']}
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


def test_create_environment_object(test_environment):
    """Creates an Environment objects and queries its properties."""
    env = test_environment
    assert env.name == 'example-environment'
    assert env.tags.get('humilis-environment') == env.name


def test_environment_not_already_in_aws(test_environment):
    """Ensures that the test environment is not already in CF."""
    test_environment.delete()
    assert not test_environment.in_cf


def test_create_environment(test_environment):
    """Creates a test environment in CF."""
    test_environment.create()
    assert test_environment.in_cf
    test_environment.delete()
    assert not test_environment.in_cf



# -*- coding: utf-8 -*-
"""Project metadata."""

package = 'humilis'
project = 'humilis'
version = '0.0.4'
description = 'Helps you deploy infrastructure in the AWS cloud'
authors = ['German Gomez-Herrero', 'Innovative Travel Ltd']
authors_string = ', '.join(authors)
emails = ['german@innovativetravel.eu']
license = 'MIT'
copyright = '2015 ' + authors_string
url = 'http://github.com/InnovativeTravel/humilis'



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import os.path
import re
import logging
from humilis.config import config
from humilis.utils import DirTreeBackedObject, get_cf_name
from humilis.exceptions import ReferenceError, CloudformationError
from boto3facade.s3 import S3
from boto3facade.ec2 import Ec2
from boto3facade.exceptions import NoUpdatesError
import json
import time
import datetime


class Layer:
    """A layer of infrastructure that translates into a single CF stack"""
    def __init__(self, environment, name, logger=None, loader=None,
                 **user_params):
        self.__environment_repr = repr(environment)
        self.cf = environment.cf
        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger
        self.name = name
        self.env_name = environment.name
        self.env_stage = environment.stage
        self.env_basedir = environment.basedir
        self.depends_on = []
        self.section = {}
        if loader is None:
            loader = DirTreeBackedObject(self.basedir, self.logger)
        self.loader = loader
        self.children = []

        # These param set will be sent to the template compiler and will be
        # populated once the layers this layer depend on have been created.
        self.params = {}

        self.meta = self.loader.load_section('meta', params=self.loader_params)
        for dep in self.meta.get('dependencies', []):
            dep_cf_name = get_cf_name(self.env_name, dep, stage=self.env_stage)
            self.depends_on.append(dep_cf_name)

        self.sns_topic_arn = environment.sns_topic_arn
        self.tags = {
            'humilis-layer': self.name,
            'humilis-created-on': str(datetime.datetime.now())}
        for tagname, tagvalue in environment.tags.items():
            self.tags[tagname] = tagvalue

        for tagname, tagvalue in self.meta.get('tags', {}).items():
            self.tags[tagname] = tagvalue

        self.yaml_params = self.meta.get('parameters', {})

        # User params override what is in the layer definition file
        self.user_params = user_params
        for pname, pvalue in user_params.items():
            if pname not in self.yaml_params:
                msg = "Unknown parameter {pname} for layer {layer}: ignored"\
                    .format(pname=pname, layer=self.name)
                self.logger.warning(msg)
            else:
                self.yaml_params[pname]['value'] = pvalue
        self.__ec2 = None
        self.__s3 = None

    @property
    def cf_name(self):
        """The name of the CF stack associated to this layer."""
        return get_cf_name(self.env_name, self.name, stage=self.env_stage)

    @property
    def loader_params(self):
        """Produces a dictionary of parameters to pass to a section loader."""
        # User parameters in the layer meta.yaml
        params = {k: v['value'] for k, v in self.params.items()}
        params['_env'] = {'stage': self.env_stage, 'name': self.env_name}
        params['_os_env'] = os.environ
        params['_layer'] = {'name': self.name}
        return params

    @property
    def basedir(self):
        return os.path.join(self.env_basedir, 'layers', self.name)

    @property
    def in_cf(self):
        """Returns true if the layer has been already deployed to CF."""
        return self.cf_name in {stk['StackName'] for stk in self.cf.stacks}

    @property
    def ec2(self):
        """Connection to AWS EC2 service."""
        if self.__ec2 is None:
            self.__ec2 = Ec2(config.boto_config)
        return self.__ec2

    @property
    def s3(self):
        """Connection to AWS S3."""
        if self.__s3 is None:
            self.__s3 = S3(config.boto_config)
        return self.__s3

    @property
    def children_in_cf(self):
        """List of (already created) children layers."""
        return [x for x in self.children if x.in_cf]

    @property
    def ok(self):
        """Layer is fully deployed in CF and ready for use"""
        return self.cf.stack_ok(self.cf_name)

    @property
    def outputs(self):
        """Layer outputs. Throws an exception if layer is not ok."""
        if not self.ok:
            msg = ("Attempting to read outputs from a layer that has not "
                   "been fully deployed yet")
            raise CloudformationError(msg, logger=self.logger)
        ly = self.cf.stack_outputs[self.cf_name]
        if ly is not None:
            ly = {o['OutputKey']: o['OutputValue'] for o in ly}
        return ly

    @property
    def dependencies_met(self):
        """Checks whether all stack dependencies have been deployed."""
        current_cf_stack_names = {stack.get('StackName') for stack
                                  in self.cf.stacks}
        for dep in self.depends_on:
            if dep not in current_cf_stack_names:
                return False
        return True

    def add_child(self, child):
        """Adds a child layer to this layer."""
        self.children.append(child)
        # Don't use a comma as a separator. For whatever reason CF seems to
        # occasionally break when tag values contain commas.
        self.tags['humilis-children'] = ':'.join((x.name for x
                                                  in self.children))

    def compile(self):
        """Loads all files associated to a layer."""
        # Some templates may refer to params, so populate them first
        self.populate_params()

        # Load all files with layer contents
        for section in config.LAYER_SECTIONS:
            self.section[section] = self.loader.load_section(
                section, params=self.loader_params)

        # Package the layer as a CF template
        cf_template = {
            'AWSTemplateFormatVersion': str(config.CF_TEMPLATE_VERSION),
            'Description': self.meta.get('description', ''),
            'Mappings': self.section.get('mappings', {}),
            'Parameters': self.section.get('parameters', {}),
            'Resources': self.section.get('resources', {}),
            'Outputs': self.section.get('outputs', {})
        }
        return cf_template

    def populate_params(self):
        """Populates parameters in a layer by resolving references."""
        if len(self.yaml_params) < 1:
            return
        for pname, param in self.yaml_params.items():
            self.params[pname] = {}
            self.params[pname]['description'] = param.get('description', None)
            self.params[pname]['value'] = self._parse_param_value(
                param['value'])

    def print_params(self):
        """Prints the params used during layer creation."""
        if len(self.params) < 1:
            print("No parameters. Did you forget to run populate_params()?")
            return
        print("Parameters for layer {}:".format(self.cf_name))
        for pname, param in self.params.items():
            pval = param.get('value', None)
            if len(pval) > 30:
                pval = pval[0:30]
            print("{pname:<15}: {pval:>30}".format(pname=pname, pval=pval))

    def _parse_param_value(self, pval):
        """Parses layer parameter values."""
        if isinstance(pval, list):
            # A list of values: parse each one individually
            return [self._parse_param_value(_) for _ in pval]
        elif isinstance(pval, dict) and 'ref' in pval:
            # A reference
            return self._resolve_ref(pval['ref'])
        elif isinstance(pval, dict):
            return {k: self._parse_param_value(v) for k, v in pval.items()}
        else:
            return pval

    def _resolve_ref(self, ref):
        """Resolves references."""
        parser = config.reference_parsers.get(ref.get('parser'))
        if parser is None:
            msg = "Invalid reference in layer {}".format(self.cf_name)
            raise ReferenceError(ref, msg, logger=self.logger)
        parameters = ref.get('parameters', {})
        result = parser(self, config.boto_config, **parameters)
        return result

    def delete(self):
        """Deletes a stack in CF."""
        msg = "Deleting stack {} from CF".format(self.cf_name)
        self.logger.info(msg)
        if len(self.children_in_cf) > 0:
            msg = "Layer {} has dependencies ({}): will not be deleted".\
                format(self.name, [x.name for x in self.children_in_cf])
            self.logger.info(msg)
        else:
            self.cf.delete_stack(self.cf_name)

    def create(self, update=False):
        """Deploys a layer as a CF stack."""
        msg = "Starting checks for layer {}".format(self.name)
        self.logger.info(msg)

        if not self.dependencies_met:
            msg = "Dependencies for layer {layer} are not met, skipping"\
                .format(layer=self.name)
            self.logger.critical(msg)
            return

        cf_template = self.compile()
        # CAPABILITY_IAM is needed only for layers that contain certain
        # resources, but we add it  always for simplicity.
        if not self.in_cf:
            self.logger.info("Creating layer '{}' (CF stack '{}')".format(
                self.name, self.cf_name))

            self.cf.create_stack(
                self.cf_name,
                json.dumps(cf_template, indent=4),
                self.sns_topic_arn,
                self.tags)
        elif update:
            self.logger.info("Updating layer '{}'".format(self.name))
            try:
                self.cf.update_stack(
                    self.cf_name,
                    json.dumps(cf_template, indent=4),
                    self.sns_topic_arn)
            except NoUpdatesError:
                msg = "No updates on layer '{}'".format(self.name)
                self.logger.warning(msg)
        else:
            msg = "Layer '{}' already in CF: not creating".format(self.name)
            self.logger.info(msg)

        status = self.watch_events()
        if status not in {'CREATE_COMPLETE', 'UPDATE_COMPLETE',
                          'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS'}:
            msg = "Unable to deploy layer {}: status is {}".format(
                self.name, status)
            raise CloudformationError(msg, logger=self.logger)

        return self.outputs

    def watch_events(self, progress_status={'CREATE_IN_PROGRESS',
                                            'UPDATE_IN_PROGRESS'}):
        """Watches CF events during stack creation."""
        if not self.in_cf:
            self.logger.warning("Layer {} has not been deployed to CF: "
                                "nothing to watch".format(self.name))
            return
        stack_status = self.cf.get_stack_status(self.cf_name)
        already_seen = set()
        cm = config.EVENT_STATUS_COLOR_MAP
        while stack_status in progress_status:
            events = self.cf.get_stack_events(self.cf_name)
            new_events = [ev for ev in events if ev.id not in already_seen]
            for event in new_events:
                self.logger.info(
                    "{time} {color}{status}\033[0m {restype} {logid} {physid} "
                    "{reason}".format(
                        time=event.timestamp.isoformat(),
                        color=cm.get(event.resource_status, ''),
                        status=event.resource_status,
                        restype=event.resource_type,
                        logid=event.logical_resource_id,
                        physid=event.physical_resource_id,
                        reason=event.resource_status_reason,
                    ))
                already_seen.add(event.id)

            stack_status = self.cf.get_stack_status(self.cf_name)
            time.sleep(3)
        return stack_status

    def __repr__(self):
        return str(self)

    def __str__(self):
        args = re.sub(r'\'(\w+)\'\s*:\s*', r'\1=', str(self.user_params))[1:-1]
        if len(args) > 0:
            basestr = "Layer({env}, '{name}', {args})"
        else:
            basestr = "Layer({env}, '{name}')"
        return basestr.format(
            env=self.__environment_repr, name=self.name, args=args)
